"""
Utility functions for accessing a logged-in connection to GitHub (with
consistent handling of passwords and tokens).
"""

import functools
import logging
import netrc
import os.path

from appdirs import user_config_dir
import click
from github3 import login, GitHubError
from six.moves import input
import yaml


logging.basicConfig()
LOGGER = logging.getLogger(__name__)

CONFIG_DIR = user_config_dir('edx-repo-tools', 'edx')
AUTH_CONFIG_FILE = os.path.join(CONFIG_DIR, 'auth.yaml')

AUTHORIZATION_NOTE = 'edx-repo-tools'

TWO_FACTOR_CODE = None


def do_two_factor():
    """
    Capture two-factor auth input for use by the GitHub object.
    """
    global TWO_FACTOR_CODE  # pylint: disable=global-statement
    if TWO_FACTOR_CODE is None:
        TWO_FACTOR_CODE = input('Two-factor code: ')

    return TWO_FACTOR_CODE


def login_github(username=None, password=None, token=None, token_file=None):
    """
    Log in to GitHub using the specified username, password, token, or
    token_file.

    The token_file is used preferentially, containing a personal access token.
    If not specified, read from an auth.yaml file in the user settings
    directory.  if that doesn't exist, read ~/.netrc.  If that doesn't exist,
    prompt for username and password, create a token, and store it in the user
    settings directory.

    Arguments:
        username (string):
            The user to log in as. If not specified, checks the AUTH_SETTINGS
            dictionary.
        password (string):
            The password to log in with. If not specified, and no ``token`` is
            specified, then prompts the user.
        token (string):
            The personal access token to log in with. If not specified, checks
            the AUTH_SETTINGS dictionary.
        token_file (string):
            File containing the personal access token to log in with. Overrides
            token argument. File can be a named pipe (bash process
            substitution) for increased security.

    Returns: (:class:`~github3.GitHub`)
        A logged-in `~github3.GitHub` instance
    """
    hub = None

    try:
        with open(AUTH_CONFIG_FILE) as auth_config:
            AUTH_SETTINGS = yaml.safe_load(auth_config)
        LOGGER.info("Read auth from {!r}".format(AUTH_CONFIG_FILE))
    except:  # pylint: disable=bare-except
        LOGGER.debug('Unable to load auth settings', exc_info=True)
        AUTH_SETTINGS = {}

    if username is None:
        username = AUTH_SETTINGS.get('username')

    if token is None:
        token = AUTH_SETTINGS.get('token')

    # Log in with token from file, if it's supplied
    if token_file is not None and username is not None:
        with open(token_file, 'r') as tf:
            token = tf.readline()[:-1]
            if token:
                hub = login(username, token)
            else:
                LOGGER.warn("No token in file")

    # Otherwise, fall back to password, if supplied
    elif password is not None and username == AUTH_SETTINGS.get('username'):
        hub = login(username, password, two_factor_callback=do_two_factor)

    # Otherwise, log in with the stored token
    elif username is not None and token is not None:
        hub = login(username, token)

    else:
        # Try .netrc
        try:
            netrc_data = netrc.netrc()
        except IOError:
            # No .netrc file, that's fine.
            pass
        else:
            LOGGER.info("Read .netrc for auth")
            authenticator = netrc_data.authenticators("api.github.com")
            if authenticator is not None:
                username, _, token = authenticator
            hub = login(username, token)

    # If no password or token, prompt for a password
    # and generate a token, and then store the token
    if hub is None:
        if username is None:
            username = click.prompt('Username')

        password = click.prompt('Password', hide_input=True)

        hub = login(username, password, two_factor_callback=do_two_factor)

        try:
            token = hub.authorize(
                login=username,
                password=password,
                note=AUTHORIZATION_NOTE,
            )
        except GitHubError as exc:
            if exc.msg != "Validation Failed":
                raise

            LOGGER.debug('Attempting to delete existing authorization')

            authorizations = hub.authorizations()
            for authorization in authorizations:
                if authorization.note == AUTHORIZATION_NOTE:
                    authorization.delete()

            token = hub.authorize(
                login=username,
                password=password,
                scopes=['repo'],
                note=AUTHORIZATION_NOTE,
            )

        if not os.path.exists(CONFIG_DIR):
            os.makedirs(CONFIG_DIR)

        # pylint: disable=redefined-outer-name
        with open(AUTH_CONFIG_FILE, 'w') as auth_config:
            yaml.safe_dump({
                'username': username,
                'token': token.token,
            }, auth_config)
            LOGGER.info("Wrote credentials to {!r}".format(AUTH_CONFIG_FILE))

    hub.set_user_agent(AUTHORIZATION_NOTE)

    LOGGER.debug('Rate limit remaining: %d', hub.ratelimit_remaining)

    return hub


def pass_github(f):
    """
    A click decorator that passes a logged-in GitHub instance to a click
    interface (and exposes the appropriate arguments to configure that
    instance).

    For example:

        @click.command()
        @pass_github
        @click.option(
            '--dry/--yes',
            default=True,
            help='Actually create the pull requests',
        )
        def explode(hub, dry):
            hub.organization('edx').repositories()
    """

    # Mark that pass_github has been applied already to
    # `f`, so that if the decorator is applied multiple times,
    # it won't pass the `hub` argument multiple times, and
    # so that multiple copies of the click arguments won't be added.
    if getattr(f, '_pass_github_applied', False):
        return f
    f._pass_github_applied = True

    # pylint: disable=missing-docstring
    @click.option(
        '--username',
        help='Specify the user to log in to GitHub with',
    )
    @click.option('--password', help='Password to log in to GitHub with')
    @click.option(
        '--token',
        help='Personal access token to log in to GitHub with',
    )
    @click.option(
        '--token-file',
        help='File containing personal access token to log in to GitHub with',
    )
    @click.option(
        '--debug/--no-debug',
        help='Enable debug logging',
        default=False
    )
    @functools.wraps(f)
    def wrapped(username, password, token, token_file, debug, *args, **kwargs):

        if debug:
            logging.basicConfig()
            logging.getLogger().setLevel(logging.DEBUG)

        hub = login_github(username, password, token, token_file)

        f(hub=hub, *args, **kwargs)

    return wrapped



import io
import re
from configparser import ConfigParser, NoSectionError

import click

TOX_SECTION = "tox"
ENVLIST = "envlist"
TEST_ENV_SECTION = "testenv"
TEST_ENV_DEPS = "deps"
PYTHON_SUBSTITUTE = "py38"
DJANGO_SUBSTITUTE = "django{22,30,31}"

DJANGO_22_DEPENDENCY = "django22: Django>=2.2,<2.3\n"
DJANGO_30_DEPENDENCY = "django30: Django>=3.0,<3.1\n"
DJANGO_31_DEPENDENCY = "django31: Django>=3.1,<3.2\n"
NEW_DJANGO_DEPENDENCIES = DJANGO_22_DEPENDENCY + DJANGO_30_DEPENDENCY + DJANGO_31_DEPENDENCY

SECTIONS = [TOX_SECTION, TEST_ENV_SECTION]

PYTHON_PATTERN = "(py{.*?}-?|py[0-9]+,|py[0-9]+-)"

DJANGO_PATTERN = "(django[0-9]+,|django[0-9]+\n|django{.*}\n|django{.*?}|django[0-9]+-|django{.*}-)"

DJANGO_DEPENDENCY_PATTERN = "([^\n]*django[0-9]+:.*\n?)"


class ConfigReader:
    def __init__(self, file_path=None, config_dict=None):
        self.config_dict = config_dict
        self.file_path = file_path

    def get_modernizer(self):
        config_parser = ConfigParser()
        if self.config_dict is not None:
            config_parser.read_dict(self.config_dict)
        else:
            config_parser.read(self.file_path)
        return ToxModernizer(config_parser, self.file_path)


class ToxModernizer:
    def __init__(self, config_parser, file_path):
        self.file_path = file_path
        self.config_parser = config_parser
        self._validate_tox_config_sections()

    def _validate_tox_config_sections(self):
        if not self.config_parser.sections():
            raise NoSectionError("Bad Config. No sections found.")

        if all(section not in SECTIONS for section in self.config_parser.sections()):
            raise NoSectionError("File doesn't contain required sections")

    def _update_env_list(self):
        tox_section = self.config_parser[TOX_SECTION]
        env_list = tox_section[ENVLIST]

        env_list = ToxModernizer._replace_runners(PYTHON_PATTERN, PYTHON_SUBSTITUTE, env_list)
        env_list = ToxModernizer._replace_runners(DJANGO_PATTERN, DJANGO_SUBSTITUTE, env_list)
        self.config_parser[TOX_SECTION][ENVLIST] = env_list

    @staticmethod
    def _replace_runners(pattern, substitute, env_list):
        matches = re.findall(pattern, env_list)
        if not matches:
            return env_list
        substitute = ToxModernizer._get_runner_substitute(matches, substitute)
        return ToxModernizer._replace_matches(pattern, substitute, env_list, matches)

    @staticmethod
    def _replace_matches(pattern, substitute, target, matches):
        if not matches:
            return target
        occurrences_to_replace = len(matches) - 1
        if occurrences_to_replace > 0:
            target = re.sub(pattern, '', target, occurrences_to_replace)
        target = re.sub(pattern, substitute, target)
        return target

    @staticmethod
    def _get_runner_substitute(matches, substitute):
        last_match = matches[-1]
        has_other_runners = last_match.endswith('-')
        return substitute + "-" if has_other_runners else substitute

    def _replace_django_versions(self):
        test_environment = self.config_parser[TEST_ENV_SECTION]
        dependencies = test_environment[TEST_ENV_DEPS]
        matches = re.findall(DJANGO_DEPENDENCY_PATTERN, dependencies)
        dependencies = self._replace_matches(DJANGO_DEPENDENCY_PATTERN, NEW_DJANGO_DEPENDENCIES, dependencies, matches)

        self.config_parser[TEST_ENV_SECTION][TEST_ENV_DEPS] = dependencies

    def _update_config_file(self):
        # ConfigParser insists on using tabs for output. We want spaces.
        with io.StringIO() as configw:
            self.config_parser.write(configw)
            new_ini = configw.getvalue()
        new_ini = new_ini.replace("\t", "    ")
        with open(self.file_path, 'w') as configfile:
            configfile.write(new_ini)

    def modernize(self):
        self._update_env_list()
        self._replace_django_versions()
        self._update_config_file()


@click.command()
@click.option(
    '--path', default='tox.ini',
    help="Path to target tox config file")
def main(path):
    modernizer = ConfigReader(path).get_modernizer()
    modernizer.modernize()


if __name__ == '__main__':
    main()



import click

from edx_repo_tools.utils import YamlLoader

DEPRECATED_FIELDS = ['owner', 'supporting_teams', 'track_pulls', 'track-pulls']


class YamlModernizer(YamlLoader):
    """
    Yaml Modernizer class is responsible for getting rid of obsolete elements from openedx.yaml files
    update the DEPRECATED_FIELDS list to adjust the modernizer output
    """

    def __init__(self, file_path):
        super().__init__(file_path)

    def _remove_deprecated_elements(self):
        for deprecated_field in DEPRECATED_FIELDS:
            if deprecated_field in self.elements.keys():
                del self.elements[deprecated_field]

    def modernize(self):
        self._remove_deprecated_elements()
        self.update_yml_file()


@click.command()
@click.option(
    '--path', default='openedx.yaml',
    help="Path to target openedx.yaml file")
def main(path):
    modernizer = YamlModernizer(path)
    modernizer.modernize()


if __name__ == "__main__":
    main()



import sys

from typing import Optional
from bowler import Query, LN, Capture, Filename, TOKEN, SYMBOL
from fissix.pytree import Node, Leaf


# usage of the User.is_authenticated() and User.is_anonymous() methods is deprecated. Remove the last parenthesis.


def parentheses(node: LN, capture: Capture, filename: Filename) -> Optional[LN]:

    # this is how the last node look a like, Node(trailer, [Leaf(7, '('), Leaf(8, ')')])
    nodes = capture.get("function_call")
    last_node = nodes.children[-1]  # pick the last node

    # make sure its a leaf and contains parenthesis before deleting the last node
    if isinstance(last_node.children[0], Leaf) and isinstance(last_node.children[1], Leaf):
        if last_node.children[0].value == "(" and last_node.children[1].value == ")":
            del nodes.children[-1]

    return nodes


(
    Query(sys.argv[1])
    .select_method("is_authenticated")
    .modify(parentheses)
    .idiff()
),
(
    Query(sys.argv[1])
    .select_method("is_anonymous")
    .modify(parentheses)
    .idiff()
)



#!/usr/bin/env python
"""Find repos with no openedx.yaml file."""

from __future__ import print_function

import itertools

import click
from github3.exceptions import NotFoundError

from edx_repo_tools.auth import pass_github
from edx_repo_tools.data import pass_repo_tools_data, iter_nonforks
from edx_repo_tools.utils import dry_echo, dry


@click.command()
@pass_github
@pass_repo_tools_data
@click.option('--org', multiple=True, default=['edx', 'edx-ops'])
@dry
def no_yaml(hub, repo_tools_data, org, dry):
    """Find public repos with no openedx.yaml file."""
    repos = iter_nonforks(hub, org)
    repos = sorted(repos, reverse=True, key=lambda r: r.pushed_at)
    for repo in repos:
        if repo.private:
            continue
        if repo.archived:
            continue
        try:
            contents = repo.file_contents('openedx.yaml')
        except NotFoundError:
            contents = None

        if contents is None:
            print(u"{}: {}".format(repo.full_name, repo.pushed_at))
            try:
                commits = list(itertools.islice(repo.commits(), 5))
            except Exception as exc:
                print("  Error: {}".format(exc))
                continue
            for commit in commits:
                commit = commit.refresh()
                # last_modified is a string?? 'Fri, 06 Apr 2018 19:48:42 GMT'
                when = " ".join(commit.last_modified.split()[1:4])
                message = commit.commit['message'].splitlines()[0]
                who = commit.commit['author']['name']
                print(u"  {}: {} ({})".format(who, message, when))



#!/usr/bin/env python
"""
Tag repos for an Open edX release. When run, this script will:

1. Make sure we have an OAuth token for the GitHub API, and help the user
   create one if we don't already have one
2. Fetch the openedx.yaml files from all repos covered by the Open edX release.
3. Identify the repos in that file that are tagged in the Open edX release
4. Identify the commit that needs to be tagged in each repo, tracing cross-repo
   dependencies as necessary
5. Show the identified repos and commits, and ask for confirmation from the user
6. Upon confirmation, create Git tags for the repos using the GitHub API
"""

import collections
import copy
import datetime
import fnmatch
import logging

import click
from github3 import GitHubError
from github3.exceptions import NotFoundError
from tqdm import tqdm

from edx_repo_tools.auth import pass_github
from edx_repo_tools.data import iter_openedx_yaml
from edx_repo_tools.utils import dry, dry_echo

log = logging.getLogger(__name__)


# Name used for fetching/storing GitHub OAuth tokens on disk
TOKEN_NAME = "openedx-release"


# An object to act like a response (with a .text attribute) in the case that
# create_ref uselessly returns us None on failure.
FakeResponse = collections.namedtuple("FakeResponse", "text")


class TagReleaseError(Exception):
    """Something went wrong..."""
    pass


def nice_tqdm(iterable, desc):
    return tqdm(iterable, desc=desc.ljust(27))


def openedx_release_repos(hub, orgs=None, branches=None):
    """
    Return a subset of the repos with openedx.yaml files: the repos
    with an `openedx-release` section.

    Arguments:
        hub (:class:`~github3.GitHub`): an authenticated GitHub instance.
        orgs (list of str): The GitHub organizations to scan. Defaults to edx, edx-ops, and
              edx-solutions.
        branches (list of str): The branches to scan in all repos in the selected
                  orgs, defaulting to the repo's default branch.

    Returns:
        A dict from :class:`~github3.Repository` objects to openedx.yaml data for all of the
        repos with an ``openedx-release`` key specified.

    """
    if not orgs:
        orgs = ['edx', 'edx-ops', 'edx-solutions']

    repos = {}

    for repo, data in tqdm(iter_openedx_yaml(hub, orgs=orgs, branches=branches), desc='Find repos'):
        if data.get('openedx-release'):
            repo = repo.refresh()
            repos[repo] = data

    return repos


def repo_matches(repo, pattern):
    """Match a repo against a filename pattern.

    True if either the name or the full_name of the repo matches.
    """
    return (
        fnmatch.fnmatch(repo.full_name, pattern) or
        fnmatch.fnmatch(repo.name, pattern)
    )

def trim_skipped_repos(repos, skip_repos):
    """Remove repos we've been told to skip.

    Arguments:
        repos (dict): A dict mapping Repository objects to openedx.yaml data.
        skip_repos (list of str): Filename wildcard patterns for the names or
            full_names of repos we don't want.

    Returns:
        A dict similar to `repos`, but without the skipped repos.

    """
    trimmed = {}
    for repo, data in repos.items():
        if any(repo_matches(repo, pat) for pat in skip_repos):
            log.warning("Skipping {} by pattern".format(repo))
            continue
        trimmed[repo] = data
    return trimmed

def trim_dependent_repos(repos):
    """Remove dependent repos (an obsolete feature of this program).

    Repos with 'parent-repo' in their 'openedx-release' data are removed from
    the `repos` dict.  A new dict of repos is returned.

    """
    trimmed = {}

    for r, data in repos.items():
        if 'parent-repo' in data['openedx-release']:
            msg = u"Repo {repo} is dependent: you can remove openedx-release from its openedx.yaml file".format(repo=r)
            click.secho(msg, fg='yellow')
        else:
            trimmed[r] = data

    return trimmed


def trim_indecisive_repos(repos):
    """
    Check the repos for a "maybe" value in the "openedx-release" key.

    Arguments:
        repos (dict): A dict mapping Repository objects to openedx.yaml data.

    Returns:
        A dict like its argument, but without the indecisive repos.

    """
    trimmed = {}
    for repo, repo_data in repos.items():
        maybe = repo_data["openedx-release"].get("maybe")
        if maybe:
            click.secho("*** {repo} has openedx-release 'maybe', skipped".format(repo=repo), fg="red")
        else:
            trimmed[repo] = repo_data
    return trimmed


def override_repo_refs(repos, override_ref=None, overrides=None):
    """
    Apply ref overrides to the `repos` dictionary.

    Arguments:
        repos (dict): A dict mapping Repository objects to openedx.yaml data.
        override_ref (str): a ref to use in all repos.
        overrides (dict mapping repo names to refs): refs to use in specific repos.

    Returns:
        A new dict mapping Repository objects to openedx.yaml data, with refs overridden.

    """
    repos = {r: copy.deepcopy(data) for r, data in repos.items()}
    overrides = overrides or {}
    if override_ref or overrides:
        for repo, repo_data in repos.items():
            local_override = overrides.get(repo.full_name, override_ref)
            if local_override:
                repo_data["openedx-release"]["ref"] = local_override
    return repos


def commit_ref_info(repos, skip_invalid=False):
    """
    Returns a dict of information about what commit should be tagged in each repo.

    If the information in the passed-in dictionary is invalid in any way,
    this function will throw an error unless `skip_invalid` is set to True,
    in which case the invalid information will simply be logged and ignored.

    Arguments:
        repos (dict): A dict mapping Repository objects to openedx.yaml data.
        skip_invalid (bool): if true, log invalid data in `repos`, but keep going.

    Returns:
        A dict mapping Repositories to a dict about the ref to tag, like this::

        {
            Repository(<full_repo_name>): {
                "ref": "name of tag or branch"
                "ref_type": "tag", # or "branch"
                "sha": "1234566789abcdef",
                "message": "The commit message"
                "author": {
                    "name": "author's name",
                    "email": "author's email"
                }
                "committer": {
                    "name": "committer's name",
                    "email": "committer's email",
                }
            },
            Repository(<next_repo_name>): {...},
            ...
        }

    """

    ref_info = {}
    for repo, repo_data in nice_tqdm(repos.items(), desc='Find commits'):
        # are we specifying a ref?
        ref = repo_data["openedx-release"].get("ref")
        if ref:
            try:
                ref_info[repo] = get_latest_commit_for_ref(repo, ref)
            except (GitHubError, ValueError):
                if skip_invalid:
                    msg = u"Invalid ref {ref} in repo {repo}".format(
                        ref=ref,
                        repo=repo.full_name
                    )
                    log.error(msg)
                    continue
                else:
                    raise
    return ref_info


def get_latest_commit_for_ref(repo, ref):
    """
    Get information about the latest commit on a ref.

    Arguments:
        repo (Repository): the repo to examine.
        ref (str): the git ref to get information about.

    Returns:
        A dict with information about the commit.

    """
    # is it a branch?
    try:
        branch = repo.branch(ref)
    except NotFoundError:
        pass
    else:
        commit = repo.git_commit(branch.commit.sha).refresh()
        return {
            "ref": ref,
            "ref_type": "branch",
            "sha": commit.sha,
            "message": commit.message,
            "author": commit.author,
            "committer": commit.committer,
        }

    try:
        tag = repo.ref('tags/{}'.format(ref))
    except TypeError as err:
        # GitHub unfortunately returns a list of partial matches if you ask for
        # a ref that doesn't exist.  This means the code in github3 that
        # expects a dict (the ref that matches) actually gets a list instead.
        # Then it tries to pop, and a TypeError occurs
        # https://github.com/sigmavirus24/github3.py/issues/310
        # We'll catch the error and make the problem clearer.
        if "pop() takes at most 1 argument (2 given)" in str(err):
            raise ValueError(u"In repo {}, ref {!r} doesn't exist.".format(repo, ref))
        else:
            raise
    except NotFoundError:
        try:
            # Maybe it's a commit sha?
            commit = repo.git_commit(ref).refresh()
        except NotFoundError:
            raise ValueError(u"In repo {}, ref {!r} doesn't exist.".format(repo, ref))
        else:
            return {
                "ref": ref,
                "ref_type": "commit",
                "sha": ref,
                "message": commit.message,
                "author": commit.author,
                "committer": commit.committer,
            }

    if tag.object.type == "tag":
        # An annotated tag, one more level of indirection.
        tag = repo.tag(tag.object.sha)
    # need to do a subsequent API call to get the tagged commit
    commit = repo.git_commit(tag.object.sha).refresh()
    if commit:
        return {
            "ref": ref,
            "ref_type": "tag",
            "sha": commit.sha,
            "message": commit.message,
            "author": commit.author,
            "committer": commit.committer,
        }

    raise ValueError(u"No commit for {ref} in {repo}".format(ref=ref, repo=repo.full_name))



def get_ref_for_repos(repos, ref, use_tag=True):
    """
    Returns a dictionary with a key-value pairing for each repo in the given
    list of repos where the given ref exists in that repo. The key is the
    full name of the repo, and the value is a dictionary of information about
    the commit that the ref points to in that repo.
    If no repos contain the given ref, this function will return an empty dict
    -- as a result, you can use the return value of this function to check
    if the ref exists in any repos, just by coercing the return value
    to a boolean. (Empty dicts are falsy, populated dicts are truthy.)
    """

    # Github3's repo.ref function wants a ref without a leading "refs/", even
    # though the create_ref function wants one *with* a leading "refs/". :(
    if ref.startswith("refs/"):
        ref = ref[5:]
    elif not ref.startswith(("heads/", "tags/")):
        ref = "{type}/{name}".format(
            type="tags" if use_tag else "heads",
            name=ref,
        )
    return_value = {}
    for repo in nice_tqdm(repos, desc='Get refs'):
        try:
            ref_obj = repo.ref(ref)
        except NotFoundError:
            pass
        except TypeError:
            # If the ref isn't found, GitHub uses the ref as a substring,
            # and returns all the refs that start with that string as an
            # array. That causes github3 to throw a type error when it
            # tries to pop a dict key from a list
            pass
        else:
            if ref_obj.object.type == "tag":
                # this is an annotated tag -- fetch the actual commit
                ref_obj = repo.tag(ref_obj.object.sha)
            commit = repo.git_commit(ref_obj.object.sha).refresh()
            # save the sha value for the commit into the returned dict
            return_value[repo.full_name] = {
                "ref": "refs/" + ref,
                "ref_type": "tag" if use_tag else "branch",
                "sha": commit.sha,
                "message": commit.message,
                "author": commit.author,
                "committer": commit.committer,
            }

    return return_value


def todo_list(ref_info):
    """
    Returns a string, suitable to be printed on the command line,
    that contains a record of the repos and commits that are about to be modified.
    If no ref info is passed in, return None.
    """
    if not ref_info:
        return None

    entries = []
    for repo, commit_info in ref_info.items():
        when = datetime.datetime.strptime(commit_info['committer']['date'], "%Y-%m-%dT%H:%M:%SZ")
        entries.append(u"{repo}: {ref} ({type}) {sha}\n  {when:%Y-%m-%d} {who}: {msg}".format(
            repo=repo,
            ref=commit_info['ref'],
            type=commit_info['ref_type'],
            sha=commit_info['sha'][0:7],
            msg=commit_info["message"].splitlines()[0],
            when=when,
            who=commit_info['committer']['name'],
        ))
    return "\n".join(sorted(entries))


def create_ref_for_repos(ref_info, ref, use_tag=True, rollback_on_fail=True, dry=True):
    """
    Create refs on the given repos.

    If `rollback_on_fail` is True, then on any failure, try to delete the refs
    that we're just created, so that we don't fail in a partially-completed
    state. (Note that this is *not* a reliable rollback -- other people could
    have already fetched the refs from GitHub, or the deletion attempt might
    itself fail!)

    If this function succeeds, it will return True. If this function fails,
    but the world is in a consistent state, this function will return False.
    The world is consistent if *no* refs were successfully created on repos in the first
    place, or all created refs were were successfully rolled
    back (because `rollback_on_fail` is set to True). If this function fails,
    and the world is in an inconsistent state, this function will raise a
    TagReleaseError. This could happen if some (but not all) of the refs are created,
    and either rollback is not attempted (because `rollback_on_fail` is set to
    False), or rollback fails.

    Arguments:
        ref_info (dict mapping Repositories to commit info dicts)
        ref (str): the ref to create.
        use_tag (bool): make a tag (True) or a branch (False).
        rollback_on_fail (bool)
        dry (bool): if True, don't do anything, but print what would be done.

    Returns
        True on success, False otherwise.

    """
    if not ref.startswith("refs/"):
        ref = u"refs/{type}/{ref}".format(
            type="tags" if use_tag else "heads",
            ref=ref,
        )
    succeeded = []
    failed_resp = None
    failed_repo = None
    for repo, commit_info in ref_info.items():
        try:
            dry_echo(
                dry,
                u'Creating ref {} with sha {} in repo {}'.format(
                    ref, commit_info['sha'], repo.full_name
                ),
                fg='green'
            )
            if not dry:
                created_ref = repo.create_ref(ref=ref, sha=commit_info['sha'])
                if created_ref is None:
                    failed_resp = FakeResponse(text="Something went terribly wrong, not sure what")
                    failed_repo = repo
                    break
                succeeded.append((repo, created_ref))
        except GitHubError as exc:
            failed_resp = exc.response
            failed_repo = repo
            # don't try to tag any others, just stop
            break

    if failed_resp is None:
        return True

    # if we got to this point, then there was a failure.
    try:
        original_err_msg = failed_resp.json()["message"]
    except Exception:
        original_err_msg = failed_resp.text

    if not succeeded:
        msg = (
            u"Failed to create {ref} on {failed_repo}. "
            u"Error was {orig_err}. No refs have been created on any repos."
        ).format(
            ref=ref,
            failed_repo=failed_repo.full_name,
            orig_err=original_err_msg,
        )
        log.error(msg)
        return False

    if rollback_on_fail:
        rollback_failures = []
        for repo, created_ref in succeeded:
            try:
                dry_echo(
                    dry,
                    u'Deleting ref {} from repo {}'.format(
                        created_ref.ref, repo.full_name
                    ),
                    fg='red'
                )
                if not dry:
                    created_ref.delete()
            except GitHubError as exc:
                rollback_failures.append(repo.full_name)

        if rollback_failures:
            msg = (
                u"Failed to create {ref} on {failed_repo}. "
                u"Error was {orig_err}. "
                u"Attempted to roll back, but failed to delete ref on "
                u"the following repos: {rollback_failures}"
            ).format(
                ref=ref,
                failed_repo=failed_repo.full_name,
                orig_err=original_err_msg,
                rollback_failures=", ".join(rollback_failures)
            )
            err = TagReleaseError(msg)
            err.response = failed_resp
            err.repos = rollback_failures
            raise err
        else:
            msg = (
                u"Failed to create {ref} on {failed_repo}. "
                u"Error was {orig_err}. However, all refs were successfully "
                u"rolled back."
            ).format(
                ref=ref,
                failed_repo=failed_repo.full_name,
                orig_err=original_err_msg,
            )
            log.error(msg)
            return False
    else:
        # don't try to rollback, just raise an error
        msg = (
            u"Failed to create {ref} on {failed_repo}. "
            u"Error was {orig_err}. No rollback attempted. Refs exist on "
            u"the following repos: {tagged_repos}"
        ).format(
            ref=ref,
            failed_repo=failed_repo.full_name,
            orig_err=original_err_msg,
            tagged_repos=", ".join(repo.full_name for repo, _ in succeeded)
        )
        err = TagReleaseError(msg)
        err.response = failed_resp
        err.repos = succeeded
        raise err


def remove_ref_for_repos(repos, ref, use_tag=True, dry=True):
    """
    Delete the ref `ref` from each repository in `repos`.

    If the ref does not exist on the repo, it is skipped.

    This function returns True if any repos had the reference removed,
    or False if no repos were modified. If an error occurs while trying
    to remove a ref from a repo, the function will continue trying to
    remove refs from all the other repos in the iterable -- but after all repos
    have been attempted, this function will raise a TagReleaseError with
    a list of all the repos that did not have the ref removed.
    Trying to remove a ref from a repo that does not have that ref
    to begin with is *not* treated as an error.

    Arguments:
        repos: An iterable of Repository objects.
        ref (str): the ref to remove.
        use_tag (bool): ref is a tag (True) or a branch (False).
        dry (bool): if True, don't do anything, but print what would be done.

    Returns:
        True if any repos had the ref removed, False if no repos were modified.

    """
    if ref.startswith('refs/'):
        ref = ref[len('refs/'):]

    if not (ref.startswith("heads/") or ref.startswith('tags/')):
        ref = "{type}/{ref}".format(
            type="tags" if use_tag else "heads",
            ref=ref,
        )

    failures = {}
    modified = False
    for repo in repos:
        try:
            try:
                ref_obj = repo.ref(ref)
            except NotFoundError:
                # tag didn't exist to begin with; not an error
                continue

            dry_echo(
                dry,
                u'Deleting ref {} from repo {}'.format(
                    ref_obj.ref, repo.full_name
                ),
                fg='red'
            )
            if not dry:
                ref_obj.delete()
            modified = True
        except GitHubError as err:
            # Oops, we got a failure. Record it and move on.
            failures[repo.full_name] = err

    if failures:
        msg = (
            u"Failed to remove the ref from the following repos: {repos}"
        ).format(
            repos=", ".join(failures.keys())
        )
        err = TagReleaseError(msg)
        err.failures = failures
        raise err

    return modified


def archived_repos(repos):
    """
    Check `repos`, and return the subset that are archived.

    Arguments:
        repos (dict): A dict mapping Repository objects to openedx.yaml data.

    Returns:
        A list of Repository objects that are archived.

    """
    archived = []
    for repo in nice_tqdm(repos, desc='Check for archived repos'):
        repo = repo.refresh()
        if repo.archived:
            archived.append(repo)
    return archived


def ensure_writable(repos):
    """
    Prompt the user, and wait until these repos are all unarchived.

    Arguments:
        repos: a list of Repository objects.
    """
    while repos:
        click.secho(u"The following repos need to be unarchived to continue:", fg='red', bold=True)
        for repo in repos:
            click.echo(u"  {}: https://github.com/{}/settings".format(repo.full_name, repo.full_name))
        while not click.confirm(u"Are they all unarchived?"):
            pass
        repos = archived_repos(repos)
    click.echo(u"Thanks, they will be re-archived automatically")


@click.command()
@click.argument(
    'ref', metavar="REF",
)
@click.option(
    '--tag/--branch', "use_tag", is_flag=True, default=True,
    help=u"Whether to create branches or tags in the repo. Defaults to using tags."
)
@click.option(
    '--override-ref', metavar="REF",
    help=u"A reference to use that overrides the references from the "
         u"openedx.yaml file in *ALL* repos. This might be a release candidate "
         u"branch, for example."
)
@click.option(
    '--override', 'overrides',
    nargs=2, metavar="REPO REF",
    multiple=True,
    help=u"Override a reference for a specific repo. The repo must be "
         u"specified using the full name of the repo, like 'edx/edx-platform'. "
         u"This option can be provided multiple times."
)
@click.option(
    '-y', '--yes', 'interactive', is_flag=True, default=True, flag_value=False,
    help=u"non-interactive mode: answer yes to all questions"
)
@click.option(
    '-q', '--quiet', is_flag=True, default=False,
    help=u"don't print any unnecessary output"
)
@click.option(
    '-R', '--reverse', is_flag=True, default=False,
    help=u"delete ref instead of creating it"
)
@click.option(
    '--skip-invalid', is_flag=True, default=False,
    help=u"if the openedx.yaml file points to an invalid repo, skip it "
         u"instead of throwing an error"
)
@click.option(
    '--skip-repo', 'skip_repos', multiple=True,
    help="Specify patterns of repos that should be ignored in spite of having an openedx.yaml file."
)
@click.option(
    '--search-branch', 'branches', multiple=True,
    help="Specify a branch to search for the openedx.yaml file. If specified "
         "multiple times, the first openedx.yaml file found will be used.",
)
@click.option(
    '--org', 'orgs', multiple=True, default=['edx', 'edx-ops', 'edx-solutions'],
    help="Specify a GitHub organization to search for openedx release data. "
         "May be specified multiple times.",
)
@dry
@pass_github
def main(hub, ref, use_tag, override_ref, overrides, interactive, quiet,
         reverse, skip_invalid, skip_repos, dry, orgs, branches):
    """Create/remove tags & branches on GitHub repos for Open edX releases."""

    repos = openedx_release_repos(hub, orgs, branches)
    if not repos:
        raise ValueError(u"No repos marked for openedx-release in their openedx.yaml files!")

    repos = trim_skipped_repos(repos, skip_repos)
    repos = trim_dependent_repos(repos)
    repos = trim_indecisive_repos(repos)
    repos = override_repo_refs(
        repos,
        override_ref=override_ref,
        overrides=dict(overrides or ()),
    )

    archived = archived_repos(repos.keys())
    if archived:
        if dry:
            dry_echo(dry, u"Will need to unarchive these repos: {}".format(
                ", ".join(repo.full_name for repo in archived)
                ))
        else:
            ensure_writable(archived)

    try:
        ret = do_the_work(repos, ref, use_tag, reverse, skip_invalid, interactive, quiet, dry)
    finally:
        for repo in archived:
            dry_echo(dry, u"Re-archiving {}".format(repo.full_name))
            if not dry:
                repo.edit(repo.name, archived=True)

    return ret


def do_the_work(repos, ref, use_tag, reverse, skip_invalid, interactive, quiet, dry):
    """
    The meat of the work for tag_release.

    Arguments:
        repos (dict): A dict mapping Repository objects to openedx.yaml data.
        ref (str): the ref to create.

    """
    existing_refs = get_ref_for_repos(repos, ref, use_tag=use_tag)

    if reverse:
        if not existing_refs:
            msg = (
                u"Ref {ref} is not present in any repos, cannot remove it"
            ).format(
                ref=ref,
            )
            click.echo(msg)
            return False
        if interactive or not quiet:
            click.echo(todo_list(existing_refs))
        if interactive:
            if not click.confirm(u"Remove these refs?"):
                return

        modified = remove_ref_for_repos(repos, ref, use_tag=use_tag, dry=dry)
        if not quiet:
            if modified:
                click.echo(u"Success!")
            else:
                click.echo(u"No refs modified")
        return modified

    else:
        if existing_refs:
            msg = (
                u"The {ref} ref already exists in the following repos: {repos}"
            ).format(
                ref=ref,
                repos=", ".join(existing_refs.keys()),
            )
            raise ValueError(msg)

        ref_info = commit_ref_info(repos, skip_invalid=skip_invalid)
        if interactive or not quiet:
            click.echo(todo_list(ref_info))
        if interactive:
            if not click.confirm(u"Is this correct?"):
                return
        result = create_ref_for_repos(ref_info, ref, use_tag=use_tag, dry=dry)

        if not quiet:
            if result:
                click.echo(u"Success!")
            else:
                raise ValueError(u"Failed to create refs, but rolled back successfully")
        return result



"""Clone an entire GitHub organization into current directory."""

import os.path

import click
from git.repo.base import Repo

from edx_repo_tools.auth import pass_github


@click.command()
@click.option(
    '--forks/--no-forks', is_flag=True, default=False,
    help="Should forks be included?"
)
@click.option(
    '--depth', type=int, default=0,
    help="Depth argument for git clone",
)
@click.argument(
    'org'
)
@pass_github
def main(hub, forks, depth, org):
    for repo in hub.organization(org).repositories():
        if repo.fork and not forks:
            continue
        dir_name = repo.name
        dir_name = dir_name.lstrip("-")     # avoid dirname/option confusion
        if os.path.exists(dir_name):
            continue

        print(repo.full_name)
        clone_args = {}
        if depth:
            clone_args['depth'] = depth
        Repo.clone_from(repo.ssh_url, dir_name, **clone_args)



"""
py.test plugin to support ``oep2 report``. Adds arguments to py.test,
and creates fixtures for loading git repos.
"""

from collections import defaultdict
from git.refs.remote import RemoteReference
from git.repo.base import Repo, Head
from github3.repos.repo import Repository
import cgi
import os.path
import pytest
import textwrap
import yaml

from _pytest.terminal import pytest_report_teststatus
from edx_repo_tools.auth import login_github


SYNCED = set()


class Oep2ReportPlugin(object):
    """
    A py.test plugin that wires together the fixtures needed to run the reports.
    """

    def __init__(self, hub):
        self.hub = hub
        self.config = None
        self._repos = None
        self.tests_run = set()
        self.results = defaultdict(dict)
        self.ids_used = []

    def pytest_configure(self, config):
        self.config = config

    def pytest_addoption(self, parser):
        """
        Add options to py.test
        """
        group = parser.getgroup("OEP", "OEP reporting", "general")
        group.addoption(
            "--org", action="append", default=[],
            help="list of orgs to run tests on"
        )
        group.addoption(
            "--repo", action="append", default=[],
            help="list of specific repositories (specified as org/repo) "
                 "to run tests on"
        )
        group.addoption(
            "--username", action="store", default=None,
            help="username to log into github with"
        )
        group.addoption(
            "--password", action="store", default=None,
            help="password to log into github with"
        )
        group.addoption(
            "--token", action="store", default=None,
            help="personal access token to long into github with"
        )
        group.addoption(
            "--oep", action="append", default=[3, 4, 5, 6, 7, 9, 10, 11, 12, 14],
            help="List of OEPs to check for explicit specification of compliance"
        )
        group.addoption(
            "--checkout-root", action="store", default=".oep2-workspace",
            help="Where to check out repos that are being checked for oep2 compliance",
        )

    def get_repos(self):
        """
        Log in to GitHub and retrieve all of the repos specified on the commandline.
        """

        # N.B. This is a separate function because pytest_generate_tests is called
        # for every check_* function found by py.test, and we don't want to
        # ping github for every checker to find out the list of repos that we're
        # going to test.

        # It's separate from __init__ because we don't want to attempt to login
        # whenever this plugin is loaded, just when it's asked to get a list of
        # repos to pass to a checker function.

        if self._repos is not None:
            return self._repos

        if self.config.option.repo:
            self._repos = [
                self.hub.repository(*repo.split('/'))
                for repo in self.config.option.repo
            ]
        elif self.config.option.org:
            self._repos = [
                repo
                for org in self.config.option.org
                for repo in self.hub.organization(org).repositories()
                if not repo.fork
            ]

        return self._repos

    def pytest_generate_tests(self, metafunc):
        """
        Generate test instances for all repositories to be checked.
        """
        if 'github_repo' in metafunc.fixturenames:
            if not metafunc.config.option.org and not metafunc.config.option.repo:
                metafunc.parametrize(
                    "git_repo",
                    [Repo('.')],
                )

                metafunc.parametrize(
                    "github_repo",
                    [None],
                )
            else:
                metafunc.parametrize(
                    "github_repo",
                    self.get_repos(),
                )

        if 'oep' in metafunc.fixturenames:
            metafunc.parametrize(
                "oep",
                metafunc.config.option.oep,
                ids=["OEP-{}".format(oep) for oep in metafunc.config.option.oep],
            )

    def pytest_make_parametrize_id(self, config, val):
        if isinstance(val, Repo):
            test_id = "local"
        elif isinstance(val, Repository):
            test_id = "{}/{}".format(val.owner, val.name)
        else:
            test_id = None

        if test_id:
            self.ids_used.append(test_id)
        return test_id

    @pytest.fixture()
    def git_repo(self, request, github_repo, branch=None, remote='origin', checkout_root=None):
        """
        py.test fixture to clone a GitHub based repo onto the local disk.

        Arguments:
            github_repo (:class:`~github3.GitHub`): The repo to read from
            branch (str): The branch to check out

        Returns:
            A :class:`~git.repo.base.Repo` object, with the master branch checked out
            and up to date with the remote.
        """
        if checkout_root is None:
            checkout_root = request.config.option.checkout_root

        if not os.path.exists(checkout_root):
            os.makedirs(checkout_root)

        repo_dir = os.path.join(
            os.path.join(checkout_root, github_repo.owner.name),
            github_repo.name
        )

        if github_repo.private:
            repo_url = github_repo.ssh_url
        else:
            repo_url = github_repo.clone_url

        if not os.path.exists(repo_dir):
            repo = Repo.clone_from(repo_url, repo_dir)
        else:
            repo = Repo(repo_dir)

        if github_repo not in SYNCED:

            try:
                remote_obj = repo.remote(remote)
            except ValueError:
                repo.create_remote(remote, repo_url)
                remote_obj = repo.remote(remote)

            if remote_obj.fetch != repo_url:
                remote_obj.set_url(repo_url)

            remote_obj.fetch()
            SYNCED.add(github_repo)

        if branch is None:
            branch = github_repo.default_branch

        head = repo.head
        remote_branch = RemoteReference(repo, 'refs/remotes/{}/{}'.format(remote, branch))
        local_branch = Head(repo, 'refs/heads/{}'.format(branch))

        try:
            if head.commit != remote_branch.commit:
                local_branch.commit = remote_branch.commit
                local_branch.checkout()

        except ValueError:
            pytest.xfail("Branch {} is empty".format(branch))

        return repo

    @pytest.fixture()
    def openedx_yaml(self, git_repo):  # pylint: disable=redefined-outer-name
        """
        py.test fixture to read the openedx.yaml file from the supplied github_repo.

        Arguments:
            github_repo (:class:`~github3.GitHub`): The repo to read from

        Returns:
            A dictionary with the parsed contents of openedx.yaml.
        """
        try:
            with open(os.path.join(git_repo.working_tree_dir, "openedx.yaml")) as openedx_yaml_file:
                return yaml.safe_load(openedx_yaml_file)
        except IOError:
            return None

    def pytest_runtest_logreport(self, report):
        if report.when != 'call':
            return

        name, _, args = report.nodeid.partition('[')
        args = args.strip(']')
        repo = 'N/A'

        for test_id in sorted(self.ids_used, key=len, reverse=True):
            if test_id in args:
                args = args.replace(test_id, '').strip('-')
                repo = test_id
                break

        name = name.replace('edx_repo_tools/oep2/checks/', '')

        self.tests_run.add((name, args))
        self.results[repo][(name, args)] = report

    def pytest_sessionfinish(self, exitstatus):

        test_order = sorted(self.tests_run)

        def format_report(report, test_key):
            title = cgi.escape("{}[{}]".format(test_key[0], test_key[1]))
            if report is None:
                return '<td class="skipped" title="{}"/>'.format(title)
            else:
                return '<td class="{css_class}" title="{title}">{passed}</td>'.format(
                    css_class=cgi.escape(report.outcome, quote=True),
                    title=title,
                    passed=cgi.escape(pytest_report_teststatus(report)[1]),
                )

        check_results = "\n".join(
            "<tr><th>{}</th>{}</tr>".format(
                repo,
                "\n".join(format_report(tests.get(test_key), test_key) for test_key in test_order)
            ) for repo, tests in sorted(self.results.items(), key=lambda repo__: repo__[0].lower())
        )

        with open('oep2.report.html', 'w') as report:
            report.write(textwrap.dedent("""\
                <html>
                    <head>
                        <style>
                            thead > tr > th  {{
                                text-align: left;
                                padding: 4px;
                            }}
                            tbody > tr > td {{
                                text-align: center;
                                border: 1px solid #ccc;
                                padding: 0;
                                margin: 0;
                                width: 20px;
                            }}
                            tbody > tr > th {{
                                padding-right: 10px;
                                border-bottom: 1px solid #ccc;
                            }}

                            table {{
                                border-collapse: collapse;
                            }}

                            .passed {{
                                background-color: #cfffa0;
                            }}
                            .failed {{
                                background-color: #ffa293;
                            }}
                            .skipped {{
                                background-color: #cccccc;
                            }}
                        </style>
                    </head>
                    <body>
                        <table>
                            <tbody>
                                {check_results}
                            </tbody>
                        </table>
                    </body>
                </html>
            """.format(check_results=check_results)))



#!/usr/bin/env python
"""Sync labels across all repos using labels.yaml."""

from __future__ import print_function

import json

import requests
import yaml

import click
from helpers import paginated_get
from edx_repo_tools.auth import pass_github
from edx_repo_tools.data import pass_repo_tools_data, iter_openedx_yaml
from edx_repo_tools.utils import dry_echo, dry


def set_or_delete_labels(dry, repo, new_labels):

    desired_colors = {label: data['color'] for label, data in new_labels.items() if 'color' in data}
    undesired_names = {label for label, data in new_labels.items() if data.get('delete', False)}
    existing_labels = {label.name: label for label in repo.labels()}
    existing_names = set(existing_labels.keys())
    desired_names = set(desired_colors.keys())

    for label in desired_names - existing_names:
        dry_echo(
            dry,
            "Creating label '{}' ({})".format(label, desired_colors[label]),
            fg="green"
        )
        if not dry:
            new_label = repo.create_label(label, desired_colors[label])
            if new_label is None:
                click.secho("Couldn't create label!", fg='red', bold=True)

    for label in desired_names & existing_names:
        if existing_labels[label].color.lower() != desired_colors[label].lower():
            dry_echo(
                dry,
                "Updating label '{}' to {}".format(label, desired_colors[label]),
                fg="yellow"
            )
            if not dry:
                ret = existing_labels[label].update(label, desired_colors[label])
                if not ret:
                    click.secho("Couldn't update label!", fg='red', bold=True)

    for label in undesired_names & existing_names:
        dry_echo(
            dry,
            "Deleting label '{}'".format(label),
            fg="red"
        )
        if not dry:
            ret = existing_labels[label].delete()
            if not ret:
                click.secho("Couldn't delete label!", fg='red', bold=True)


@click.command()
@pass_github
@pass_repo_tools_data
@click.option('--org', multiple=True, default=['edx', 'edx-ops'], help="Update all repos with openedx.yml in this organization")
@click.option('--repo', help="Update this specific repo")
@dry
def sync_labels(hub, repo_tools_data, org, repo, dry):
    """Update the labels in repos to have consistent text and colors."""
    if repo is not None:
        repos = [(hub.repository(*repo.split('/')), None)]
    else:
        repos = sorted(iter_openedx_yaml(hub, org))
    for repo, _ in repos:
        print("Updating labels in {}".format(repo))
        set_or_delete_labels(
            dry,
            repo,
            repo_tools_data.labels
        )



import re
from copy import deepcopy

import click

from edx_repo_tools.utils import YamlLoader

ALLOWED_PYTHON_VERSIONS = '3.8'

DEPRECATED_DJANGO_VERSIONS_PATTERN = r"django111|django20|django21"
ALLOWED_DJANGO_VERSIONS_PATTERN = r"django22|django30|django31"

DJANGO_PATTERN = r"django[0-3][0-2][0-2]?"

ALLOWED_DJANGO_VERSIONS = ['django22', 'django30', 'django31']


class TravisModernizer(YamlLoader):
    def __init__(self, file_path):
        super().__init__(file_path)

    def _update_python_dict(self):
        python_versions = self.elements.get('python', None)
        if python_versions is None:
            return
        self.elements['python'] = [ALLOWED_PYTHON_VERSIONS]

    def _update_matrix_python_versions(self):
        matrix_elements = self.elements.get("matrix", {}).get("include")
        if matrix_elements is None:
            return
        has_python_matrix = any(matrix_item.get("python") is not None for matrix_item in matrix_elements)
        if not has_python_matrix:
            return
        non_python_matrix_elements = []
        python_matrix_items = []
        for matrix_element in matrix_elements:
            if 'python' not in matrix_element.keys():
                non_python_matrix_elements.append(matrix_element)
                continue
            python_matrix_item = deepcopy(matrix_element)
            python_matrix_item['python'] = ALLOWED_PYTHON_VERSIONS
            python_matrix_items.append(python_matrix_item)
            break
        self.elements["matrix"]["include"] = non_python_matrix_elements + python_matrix_items

    @staticmethod
    def _get_updated_django_matrix_items(django_matrix_item):
        updated_django_matrix_items = []
        for django_version in ALLOWED_DJANGO_VERSIONS:
            django_matrix_item_clone = deepcopy(django_matrix_item)
            django_matrix_item_clone["env"] = re.sub(DJANGO_PATTERN, django_version, django_matrix_item_clone["env"])
            updated_django_matrix_items.append(django_matrix_item_clone)
        return updated_django_matrix_items

    @staticmethod
    def _get_updated_django_envs(django_env_item):
        updated_django_env_items = []
        for django_version in ALLOWED_DJANGO_VERSIONS:
            django_env_item = re.sub(DJANGO_PATTERN, django_version, django_env_item)
            updated_django_env_items.append(django_env_item)
        return updated_django_env_items

    def _update_django_envs(self):
        env_elements = self.elements.get("env")
        if env_elements is None:
            return
        has_django_env = any(re.search(DJANGO_PATTERN, env_item) for env_item in env_elements)
        if not has_django_env:
            return
        django_env_item = [django_env_item for django_env_item in env_elements
                           if re.search(DJANGO_PATTERN, django_env_item)][0]
        non_django_env_items = [env_item for env_item in env_elements
                                if not re.search(DJANGO_PATTERN, env_item)]
        self.elements["env"] = non_django_env_items + TravisModernizer._get_updated_django_envs(django_env_item)

    def _update_django_matrix_envs(self):
        matrix_items = self.elements.get("matrix", {}).get("include", [])
        if not matrix_items:
            return
        has_django_env = any(re.search(DJANGO_PATTERN, matrix_item.get('env', '')) for matrix_item in matrix_items)
        if not has_django_env:
            return
        django_matrix_element = [matrix_item for matrix_item in matrix_items
                                 if re.search(DJANGO_PATTERN, matrix_item.get("env", ""))][0]
        non_django_matrix_items = [matrix_item for matrix_item in matrix_items
                                   if not re.search(DJANGO_PATTERN, matrix_item.get("env", ""))]
        self.elements["matrix"]["include"] = (non_django_matrix_items +
                                              TravisModernizer._get_updated_django_matrix_items(django_matrix_element))

    def _update_python_versions(self):
        self._update_python_dict()
        self._update_matrix_python_versions()

    def _update_django_versions(self):
        self._update_django_envs()
        self._update_django_matrix_envs()

    def modernize(self):
        self._update_python_versions()
        self._update_django_versions()
        self.update_yml_file()


@click.command()
@click.option(
    '--path', default='.travis.yml',
    help="Path to target travis config file")
def main(path):
    modernizer = TravisModernizer(path)
    modernizer.modernize()


if __name__ == '__main__':
    main()



"""Get the urls to every repository in a GitHub organization."""

import os.path

import click
from git.repo.base import Repo

from edx_repo_tools.auth import pass_github


@click.command()
@click.option(
    '--forks/--no-forks', is_flag=True, default=False,
    help="Should forks be included?"
)
@click.argument(
    'org'
)
@click.option(
    '--url_type', default="ssh",
    help="options: ssh or https"
)
@click.option(
    '--output_file', default="repositories.txt",
    help="where should script output urls"
)
@click.option(
    '--add_archived', is_flag=True, default=False,
    help="Do you want urls for archived repos?")
@pass_github
def main(hub, forks, org, url_type, output_file, add_archived):
    """
    Used to get the urls for all the repositories in a github organization
    """
    repositories = []
    for repo in hub.organization(org).repositories():
        if repo.fork and not forks:
            continue
        if repo.archived and not add_archived:
            continue
        if url_type == "ssh":
            repositories.append(repo.ssh_url)
        else:
            repositories.append(repo.clone_url)
    with open(output_file, 'w') as filehandle:
        for repo_url in repositories:
            filehandle.write('%s\n' % repo_url)



"""
Modernizer for Github Actions CI
"""
from copy import deepcopy
import sys

import click

from edx_repo_tools.utils import YamlLoader

TO_BE_REMOVED_PYTHON = ['3.5']
ALLOWED_PYTHON_VERSIONS = ['3.6', '3.7', '3.8']


class GithubCIModernizer(YamlLoader):
    def __init__(self, file_path):
        super().__init__(file_path)
        self.yml_instance.default_flow_style = None
        self.yml_instance.indent(mapping=2, sequence=2, offset=0)

    def _update_matrix(self):

        python_versions = list()
        matrix_elements = dict()
        section_key = None

        for key in ['build', 'tests', 'run_tests', 'run_quality']:
            if key in self.elements['jobs']:
                section_key = key
                matrix_elements = deepcopy(self.elements['jobs'][section_key]['strategy']['matrix'])

        for key, value in matrix_elements.items():
            if key == 'python-version':
                python_versions.extend(filter(
                    lambda version: version in ALLOWED_PYTHON_VERSIONS, value))
            elif key in ['include', 'exclude']:
                without_python35 = list()
                for item in value:
                    if item['python-version'] not in TO_BE_REMOVED_PYTHON:
                        without_python35.append(item)

                if len(without_python35):
                    self.elements['jobs'][section_key]['strategy']['matrix'][key] = without_python35
                else:
                    del self.elements['jobs'][section_key]['strategy']['matrix'][key]

        self.elements['jobs'][section_key]['strategy']['matrix']['python-version'] = python_versions

    def _update_python_versions(self):
        self._update_matrix()

    def modernize(self):
        self._update_python_versions()
        self.update_yml_file()


@click.command()
@click.option(
    '--path', default='.github/workflows/ci.yml',
    help="Path to default CI workflow file")
def main(path):
    modernizer = GithubCIModernizer(path)
    modernizer.modernize()


if __name__ == '__main__':
    main()



from datetime import date
import functools
import logging
import os.path

import click
from github3.exceptions import NotFoundError
from lazy import lazy
import yaml

from edx_repo_tools.auth import pass_github


logging.basicConfig()
LOGGER = logging.getLogger(__name__)
OPEN_EDX_YAML = 'openedx.yaml'


def iter_nonforks(hub, orgs):
    """Yield all the non-fork repos in a GitHub organization.

    Arguments:
        hub (:class:`~github3.GitHub`): A connection to GitHub.
        orgs (list of str): the GitHub organizations to search.

    Yields:
        Repositories (:class:`~github3.Repository`)

    """
    for org in orgs:
        for repo in hub.organization(org).repositories():
            if repo.fork:
                LOGGER.debug("Skipping %s because it is a fork", repo.full_name)
            else:
                yield repo


def iter_openedx_yaml(hub, orgs, branches=None):
    """
    Yield the data from all openedx.yaml files found in repositories in ``orgs``
    on any of ``branches``.

    Arguments:
        hub (:class:`~github3.GitHub`): A connection to GitHub.
        orgs (list of str): A GitHub organizations to search for openedx.yaml files.
        branches (list of str): Branches to search for openedx.yaml files. If
            that file exists on multiple branches, then only the contents
            of the first will be yielded.  (optional, defaults to the default
            branch in the repo).

    Yields:
        Repositories (:class:`~github3.Repository)

    """
    for repo in iter_nonforks(hub, orgs):
        for branch in (branches or [repo.default_branch]):
            try:
                contents = repo.file_contents(OPEN_EDX_YAML, ref=branch)
            except NotFoundError:
                contents = None

            if contents is not None:
                LOGGER.debug("Found openedx.yaml at %s:%s", repo.full_name, branch)
                try:
                    yield repo, yaml.safe_load(contents.decoded)
                except Exception as exc:
                    LOGGER.error("Couldn't parse openedx.yaml from %s:%s, skipping repo", repo.full_name, branch, exc_info=True)
                break


class Person(object):
    """
    A wrapper object around data parsed from people.yaml.
    """
    def __init__(self, username, name, email, agreement, email_ok=True,
                 other_emails=None, institution=None, committer=None, jira=None,
                 comments=None, expires_on=None, before=None, beta=None,
                 is_robot=None,
                ):
        self.username = username
        self.name = name
        self.email = email
        self.email_ok = email_ok
        self.other_emails = other_emails
        self.agreement = agreement
        self.institution = institution
        self.committer = committer
        self.jira = jira
        self.comments = comments
        self.expires_on = expires_on
        self.before = before
        self.beta = beta
        self.is_robot = is_robot

    @classmethod
    def from_yaml(cls, username, yaml_data):
        """
        Create a Person object from parsed yaml data.
        """
        return cls(username=username, **yaml_data)

    def associated_with(self, *institutions):
        """
        Return True if this Person is associated with an institution in
        ``institutions``.

        Arguments:
            *institutions: The institutions to check against
        """
        if self.agreement != 'institution':
            return False

        if self.expires_on and self.expires_on < date.today():
            return False

        institutions = [inst.lower() for inst in institutions]

        if self.institution and self.institution.lower() in institutions:
            return True

        return False


class RepoToolsData(object):

    def _read_file(self, filename):
        raise NotImplementedError()

    @lazy
    def labels(self):
        """
        The parsed contents of ``labels.yaml``.
        """
        return self._read_file('labels.yaml')

    @lazy
    def orgs(self):
        """
        The parsed contents of ``orgs.yaml``.
        """
        return self._read_file('orgs.yaml')

    @lazy
    def people(self):
        """
        The parsed contents of ``people.yaml``.
        """
        return {
            username: Person.from_yaml(username, data)
            for username, data
            in self._read_file('people.yaml').items()
        }


class LocalRepoToolsData(RepoToolsData):

    def __init__(self, root):
        self.root = root

    def _read_file(self, filename):
        with open(os.path.join(self.root, filename)) as file_data:
            return yaml.safe_load(file_data)


class RemoteRepoToolsData(RepoToolsData):
    def __init__(self, repo):
        self.repo = repo

    def _read_file(self, filename):
        return yaml.safe_load(self.repo.file_contents(filename).decoded)


def pass_repo_tools_data(f):
    """
    A click decorator that passes a logged-in GitHub instance to a click
    interface (and exposes the appropriate arguments to configure that
    instance), and also passes a configured RepoToolsData.

    For example:

        @click.command()
        @pass_repo_tools_data
        @click.option(
            '--dry/--yes',
            default=True,
            help='Actually create the pull requests',
        )
        def explode(hub, repo_tools_data, dry):
            print repo_tools_data.labels
    """

    # Mark that pass_repo_tools_data has been applied already to
    # `f`, so that if the decorator is applied multiple times,
    # it won't pass the `hub` argument multiple times, and
    # so that multiple copies of the click arguments won't be added.
    if getattr(f, '_pass_repo_tools_data_applied', False):
        return f
    f._pass_repo_tools_data_applied = True

    # pylint: disable=missing-docstring
    @click.option(
        '--local',
        default='../repo-tools-data',
        help='Specify the path to a local checkout of edx/repo-tools-data to use',
    )
    @click.option(
        '--remote',
        is_flag=True,
        default=False,
        envvar='REPO_TOOLS_LATEST_PEOPLE',
        help="Use data from edx/repo-tools-data, rather than a local checkout",
    )
    @pass_github
    @functools.wraps(f)
    def wrapped(hub, local, remote, *args, **kwargs):

        if remote:
            repo_tools_data = RemoteRepoToolsData(hub.repository('edx', 'repo-tools-data'))
        else:
            repo_tools_data = LocalRepoToolsData(local)

        return f(repo_tools_data=repo_tools_data, hub=hub, *args, **kwargs)
    return wrapped



from os import path

import click

FILES = [
    'requirements/constraint.txt',
    'requirements/constraints.txt',
    'requirements/pins.txt',
]


class CommonConstraint:
    """
        CommonConstraint class is responsible for adding common constraint pin in
        the constraints file of the repository
    """

    def __init__(self):
        self.comment = "# Common constraints for edx repos\n"
        self.constraint = "-c https://raw.githubusercontent.com/edx/edx-lint/master/edx_lint/files/common_constraints" \
                          ".txt\n "
        self.file = self._get_file_name()

    def _get_file_name(self):
        for file in FILES:
            if path.exists(file):
                return file

    def _read_lines(self):
        with open(self.file, 'r') as file:
            lines = file.readlines()
        return lines

    def _insert_constraint(self, lines):
        for i in range(len(lines)):
            if not lines[i].lstrip().startswith('#'):
                lines.insert(i, "\n")
                lines.insert(i + 1, self.comment)
                lines.insert(i + 2, self.constraint)
                return lines

    def _write_file(self, lines):
        with open(self.file, 'w') as file:
            file.writelines(lines)

    def update_file(self):
        lines = self._read_lines()
        lines = self._insert_constraint(lines)
        self._write_file(lines)


@click.command()
def main():
    constraint = CommonConstraint()
    constraint.update_file()


if __name__ == "__main__":
    main()



"""
Top-level definition of the ``drip`` commandline tool.
"""

import click

from edx_repo_tools.drip_survey import upload


@click.group()
def cli():
    """
    Tools for sending out a drip-fed survey.

    The survey will be sent out every FREQUENCY days to each participant,
    but will be staggered so that surveys are sent out to participants
    throughout the survey period.
    """
    pass


cli.add_command(upload.people_to_qualtrics_csv, name='qualtrics-csv')



"""
Tools for managing drip-style surveys.

Currently, this only supports translating people.yaml into a Qualtrics CSV.
"""
from backports import csv
from datetime import date
import hashlib
import io
import sys

import click
from edx_repo_tools.data import pass_repo_tools_data

NAME = 'FirstName'
EMAIL = 'Email'
WEEK = 'Week'
ASSOCIATED_WITH = 'AssociatedWith'
UNSUBSCRIBED = 'Unsubscribed'
RECIPIENT_ID = 'RecipientID'


@click.command()
@pass_repo_tools_data
@click.option('--frequency', help="The number of weeks between surveys for each contributor", type=int, default=12)
@click.option('--update', help="A Qualtrics Contact list export to update", type=click.Path(exists=True, dir_okay=False))
def people_to_qualtrics_csv(hub, repo_tools_data, frequency, update):
    """
    Print out a formatted file as expected by Qualtrics import.
    """

    if update is not None:
        with io.open(update, newline='', encoding='utf-8') as update_data:
            reader = csv.DictReader(update_data)
            initial = {
                row[EMAIL]: row
                for row in reader
            }
        fields = [field for field in reader.fieldnames if field]
    else:
        initial = {}
        fields = [NAME, EMAIL, WEEK, ASSOCIATED_WITH, UNSUBSCRIBED]

    csv_writer = csv.DictWriter(click.get_text_stream('stdout'), fieldnames=fields, extrasaction='ignore')
    csv_writer.writeheader()
    for username, person in repo_tools_data.people.iteritems():
        if person.email is None:
            continue


        hashdigest = hashlib.md5(person.email.lower()).hexdigest()

        row = initial.get(person.email, {})
        row.update({
            NAME: person.name,
            EMAIL: person.email,
            WEEK: int(hashdigest, 16) % frequency + 1,
            ASSOCIATED_WITH: 'edX' if person.associated_with('edX', 'ArbiSoft') else 'other',
        })

        if not person.email_ok:
            row[UNSUBSCRIBED] = 'true'

        csv_writer.writerow(row)




import pytest

def check_is_oep_explicitly_satisfied(openedx_yaml, oep):

    if openedx_yaml is None:
        pytest.xfail("No openedx.yaml found")

    if 'oeps' not in openedx_yaml:
        pytest.xfail("No 'oeps' key in openedx.yaml")

    oep_key = 'oep-{}'.format(oep)

    assert oep_key in openedx_yaml['oeps']

    oep_value = openedx_yaml['oeps'][oep_key]

    if isinstance(oep_value, dict):
        if not oep_value.get('applicable', True):
            pytest.xfail('OEP is not applicable')
            return

        if not oep_value['state']:
            pytest.xfail(oep_value['reason'])
    else:
        assert oep_value



from __future__ import print_function
import ast
import logging

from path import Path
import tox.config
from packaging.requirements import Requirement, InvalidRequirement
from packaging.specifiers import SpecifierSet, Specifier
from packaging.version import Version
import six

LIBRARY_REQUIRED_DJANGO_VERSIONS = {'1.8', '1.11'}
APPLICATION_ALLOWED_DJANGO_VERSIONS = {
    SpecifierSet('>=1.8,<1.9'),
    SpecifierSet('>=1.11,<2.0'),
}

DJANGO_VERSIONS = {
    Version('{}.{}.{}'.format(major, minor, patch))
    for major, minors in {
        1: {
            8: range(19),
            9: range(14),
            10: range(8),
            11: range(1)
        }
    }.items()
    for minor, patches in minors.items()
    for patch in patches
}

LOG = logging.getLogger(__name__)


def setup_call(parsed_setup_py):
    for statement in parsed_setup_py.body:
        if not isinstance(statement, ast.Expr):
            continue

        expr = statement.value
        if not isinstance(expr, ast.Call):
            continue

        if not isinstance(expr.func, ast.Name):
            continue

        if expr.func.id != 'setup':
            continue

        return expr

    return None


def uses_pbr(parsed_setup_py):
    setup = setup_call(parsed_setup_py)

    if setup is None:
        return False

    for keyword in setup.keywords:
        if keyword.arg == 'pbr' and keyword.value:
            return True

    return False


def parsed_requirements_txt(requirements_txt):
    if not requirements_txt.exists():
        print("not exists")
        return

    for line in requirements_txt.lines():
        if line.strip().startswith('-r'):
            sub_requirements = requirements_txt.parent / line.replace('-r', '').strip()
            for req in parsed_requirements_txt(sub_requirements):
                yield req
        else:
            try:
                yield Requirement(line)
            except InvalidRequirement:
                LOG.warning('Unable to parse requirement %r in %s', line, requirements_txt)


def requirement_is_django(req):
    try:
        if isinstance(req, six.string_types):
            req = Requirement(req)
        return req.name.lower() == 'django'
    except InvalidRequirement:
        return False


def requirements_txt_has_django(requirements_txt):
    for req in parsed_requirements_txt(requirements_txt):
        if requirement_is_django(req):
            return True
    return False


def setup_py_has_django(parsed_setup_py):
    setup = setup_call(parsed_setup_py)

    for keyword in setup.keywords:
        if keyword.arg == 'install_requires':
            for elt in keyword.value.elts:
                if not isinstance(elt, ast.Str):
                    continue
                if requirement_is_django(elt.s):
                    return True

    return False


def tox_tested_django_versions(tox_ini):
    tox_config = tox.config.parseconfig(['-c', str(tox_ini.abspath())])
    tested_versions = set()

    for env, config in tox_config.envconfigs.items():
        for dep in config.deps:
            if requirement_is_django(dep.name):
                req = Requirement(dep.name)
                tested_versions |= {
                    version
                    for version in ['1.8', '1.9', '1.10', '1.11', '2.0']
                    if version in req.specifier
                }
    return tested_versions


class OEP10(object):
    def check_django_versions(self, git_repo):
        """
        A repo is either a library or an application.

        If it's a library, then it should have a setup.py or a setup.cfg, and it should have a
        tox.ini file that runs its tests with at least django 1.8 and django 1.11, if
        its setup.py lists Django as a dependency.

        If it's a repository, then it should have a requirements/base.txt
        (or at least a requirements.txt) and that should pin
        either Django==1.8 or Django==1.11, if it specifies Django as a dependency.
        """
        working_dir = Path(git_repo.working_tree_dir)

        setup_py = working_dir / 'setup.py'
        setup_cfg = working_dir / 'setup.cfg'
        requirements_base_txt = working_dir / 'requirements/base.txt'
        requirements_txt = working_dir / 'requirements.txt'
        manage_py = working_dir / 'manage.py'
        tox_ini = working_dir / 'tox.ini'

        is_django_application = manage_py.exists()

        print(is_django_application)
        print(setup_py.exists())

        if requirements_base_txt.exists():
            requirements_file = requirements_base_txt
        else:
            requirements_file = requirements_txt

        print(requirements_file.exists())

        if not is_django_application and setup_py.exists():
            parsed_setup_py = ast.parse(setup_py.bytes(), 'setup.py')

            if uses_pbr(parsed_setup_py):
                has_django = requirements_txt_has_django(requirements_file)
            else:
                has_django = setup_py_has_django(parsed_setup_py)

            if not has_django:
                return

            tested_versions = tox_tested_django_versions(tox_ini)
            assert LIBRARY_REQUIRED_DJANGO_VERSIONS in tested_versions

        elif requirements_file.exists():
            django_specifier = None

            for req in parsed_requirements_txt(requirements_file):
                if requirement_is_django(req):
                    if django_specifier is None:
                        django_specifier = req.specifier
                    else:
                        django_specifier &= req.specifier

            if django_specifier is None:
                return

            largest_accepted_version = max(
                version
                for version in DJANGO_VERSIONS
                if version in django_specifier
            )

            for supported_version in APPLICATION_ALLOWED_DJANGO_VERSIONS:
                if largest_accepted_version in supported_version:
                    return
            msg = (
                "No allowed version range contained the largest allowed django " +
                "version {} in the version specifier {}".format(largest_accepted_version, django_specifier)
            )
            assert False, msg
        else:
            assert False, "Couldn't determine if repo is an application or a library"



import sys

from bowler import Query


(
    Query(sys.argv[1])
    .select_function("__unicode__")
    .rename('__str__')
    .idiff()
),
(
    Query(sys.argv[1])
    .select_method("__unicode__")
    .is_call()
    .rename('__str__')
    .idiff()
)



import sys
from lib2to3.fixer_util import Name, KeywordArg, Comma
from typing import Optional

from bowler import Query, LN, Capture, Filename, TOKEN
from fissix.pytree import Leaf


def render_has_no_renderer(node: LN, capture: Capture, filename: Filename) -> bool:

    if 'function_def' not in capture:
        return False  # This is not a function definition, no need to add argument

    arguments = capture.get("function_arguments")[0].children

    known_arguments = [arg.value for arg in arguments if arg.value in ['name', 'value', 'attrs']]
    if len(known_arguments) != 3:
        return False  # This render doesn't belong to widget

    for arg in arguments:
        if arg.type == TOKEN.NAME and arg.value == "renderer":
            return False  # This definition already has a renderer argument.

    return True


def add_renderer(node: LN, capture: Capture, filename: Filename) -> Optional[LN]:
    arguments = capture.get("function_arguments")[0]
    new_renderer_node = KeywordArg(Name(" renderer"), Name("None"))

    if isinstance(arguments, Leaf):  # Node is a leaf and so we need to replace it with a list of things we want instead
        arguments.replace([arguments.clone(), Comma(), new_renderer_node])
    else:
        arguments.append_child(Comma())
        arguments.append_child(new_renderer_node)

    return node


(
    Query(sys.argv[1])
        .select_method("render")
        .filter(render_has_no_renderer)
        .modify(add_renderer)
        .idiff()
)



"""
Pretty simple module to help move commits between git repos with broken common history, such as ones created through
a git filter subdirectory. Has some tools for speeding up activities on large repos such as limiting history search
by include / ignore directives and by date.
"""
import io
import os
import re
import hashlib
import configparser
import datetime
import pprint
from collections import OrderedDict

import click
import git
from git.util import join_path as join

# Regex to find commit shas in previously grafted commit messages
SHA_MESSAGE_TOKEN = re.compile(r'(>>)([0-9a-fA-F]{40})(<<)')


class InputException(Exception):
    """
    Generic exception for broken inputs
    """
    pass


class Grafter(object):
    """
    Does all of the work of grafting including:
    - Compiling a list of commits across both repos for the given time period
    - Mapping paths from the old repo to the new based on given tracking information
    - Creating a new branch for the graft
    - Creating commits to the new branch that mirror relevant parts of commits made on the old repo
    - Marking commits as grafted in commit messages
    """
    # When we clone_commits the newly created branch will be stored here, unless it's a dry run
    output_branch = None

    # Dict of potentially relevant commits (contain changes to files in the tracked paths) that exist in both repos for
    # the given time period. Needs to be ordered so we can iterate over them in the order the commits were made.
    candidate_commits = OrderedDict()

    # For reporting purposes we store off files changed in candidate_commits to the original repo that *may* have
    # impacted the new repo but don't exist there. Helps in debugging issues with tracked paths.
    unmatched_original_files = set()

    def __init__(self, original_repo_name, original_repo_path, branched_repo_path, max_lookback_days=180,
                 original_repo_head="master", branched_repo_head="master", tracked=None, original_ignore=None,
                 branched_ignore=None, verbose=False, dry_run=False):
        """
        :param original_repo_path: Relative or absolute path to the original repo's location on this machine
        :param branched_repo_path: Relative or absolute path to the branched repo's location on this machine
        :param max_lookback_days: How many days back to look for changes to carry over, fewer is faster
        :param original_repo_head: Branch to use in the original repo
        :param branched_repo_head: Branch to use in the branched repo
        :param tracked: Dict mapping original repo relative paths to branched repo relative paths ex:
                        {
                            'old/path/code': 'new/path/stored/elsewhere/code',
                            'old/path/file.py': 'new/path/libraries/file.py'
                        }
        :param original_ignore: List of paths inside the tracked paths to ignore commits from in the original repo
        :param branched_ignore: List of paths inside the tracked paths to ignore commits from in the branched repo
        :param verbose: Turn on / off verbose output
        :param dry_run: When True, does not create a new branch or perform commits
        """
        self.pprint = pprint.PrettyPrinter(indent=3)

        self.dry_run = dry_run
        self.verbose = verbose

        # Get the epoch time of our max lookback
        epoch = datetime.datetime.utcfromtimestamp(0)
        self.oldest_lookback_datetime = datetime.datetime.utcnow() - datetime.timedelta(days=max_lookback_days)
        self.oldest_lookback_epoch = (self.oldest_lookback_datetime - epoch).total_seconds()

        self.original_repo_name = original_repo_name

        # Set up our repos
        self.original_repo_path = original_repo_path
        self.original_repo = git.Repo(original_repo_path)

        self.branched_repo_path = branched_repo_path
        self.branched_repo = git.Repo(branched_repo_path)

        self.sanity_check_repos()

        # Set up our branches
        self.original_repo_head = original_repo_head
        self.original_repo.git.checkout(original_repo_head)

        self.branched_repo_head = branched_repo_head
        self.branched_repo.git.checkout(branched_repo_head)

        self.sanity_check_branches()

        # Set up our tracked and ignored paths
        self.tracked = tracked

        self.original_ignore = original_ignore
        self.sanity_check_paths(self.original_repo_path, self.tracked.keys(), "Tracked")
        self.sanity_check_paths(self.original_repo_path, self.original_ignore, "Ignore")

        self.branched_ignore = branched_ignore
        self.sanity_check_paths(self.branched_repo_path, self.tracked.values(), "Tracked")
        self.sanity_check_paths(self.branched_repo_path, self.branched_ignore, "Ignore")

    def sanity_check_repos(self):
        """
        Make sure the repos exist, are not empty, and are not dirty.

        :raises InputException if conditions are not met
        """
        errors = []
        for repo in (self.original_repo, self.branched_repo):
            if repo.bare:
                errors.append("ERROR- repository {} is empty!".format(repo.working_tree_dir))

            if repo.is_dirty():
                errors.append(
                    "ERROR- repository {} is dirty! Please commit any changes before running.".format(
                        repo.working_tree_dir
                    )
                )

        self.fail_on_errors(errors)

    def sanity_check_branches(self):
        """
        Make sure our branches are attached, local, and valid

        :raises InputException if conditions are not met
        """
        errors = []
        for repo in (self.original_repo, self.branched_repo):
            if repo.head.is_detached:
                errors.append("ERROR- branch head {} is detached!".format(repo.head.name))

            if repo.head.is_remote():
                errors.append("ERROR- branch head {} is remote!".format(repo.head.name))

            if not repo.head.is_valid():
                errors.append("ERROR- branch head {} is invalid!".format(repo.head.name))

        self.fail_on_errors(errors)

    def sanity_check_paths(self, repo_path, paths, path_type):
        """
        Checks existence of a path in the given repo
        :param repo_path: Relative or absolute path to the top level of the repo
        :param paths: List of paths to test
        :param path_type: String to indicate what type of path was being tested in error messages
        :raises InputException if conditions are not met
        """
        errors = []
        for path in paths:
            full_path = join(repo_path, path)

            if not os.path.exists(full_path):
                errors.append("{} path {} does not exist!".format(path_type, full_path))
        self.fail_on_errors(errors)

    @staticmethod
    def fail_on_errors(errors):
        """
        Utility method to print and raise exceptions if any errors are passed in
        :param errors: List of error strings to output
        :raises InputException if errors is not empty
        """
        if len(errors):
            err_str = "\n".join(errors)
            print(err_str)
            raise InputException(err_str)

    @staticmethod
    def get_hexdigest_from_commit(commit):
        """
        Creates a reasonably unique sha1 hash for a given commit, based on that commit's metadata. Used to identify
        the same commit across repos since their git sha's will be different.
        :param commit: gitpython Commit object
        :return: sha1 hexdigest of some of the given commit's metadata
        """
        hsh = hashlib.sha1()
        hsh.update(b"%i" % commit.committed_date)
        hsh.update(b"%i" % commit.authored_date)
        hsh.update(commit.message.encode("utf-8"))
        hsh.update(commit.author.email.encode("utf-8"))
        hsh.update(commit.committer.email.encode("utf-8"))
        return hsh.hexdigest()

    def find_candidate_commits(self):
        """
        Walks backwards through time, populating self.candidate_commits with "commits of interest" from first the
        branched repo, then the original. Uses a combination of commit metadata and commit messages from previous grafts
        to match up commits that exist in both repos, leaving us with the ability to find commits that exist only in
        the original repo which we want to graft onto the new repo.
        """
        # If we've previously grafted a commit we'll pull the old sha out of the message and use it to prevent
        # duplicate graftings.
        found_grafted_commits = {}

        for repo, paths, repo_name in (
                (self.branched_repo, self.tracked.values(), "branched"),
                (self.original_repo, self.tracked.keys(), "original"),
        ):
            # Key into the self.candidate_commits[digest] dict so we can make this chunk of code reusable across
            # original and branched repos.
            commit_key = "{}_commit".format(repo_name)
            paths = list(paths)

            for commit in repo.commits(paths=paths):
                self.vprint("Processing commit: {}".format(commit.hexsha))
                digest = self.get_hexdigest_from_commit(commit)

                if commit.committed_date < self.oldest_lookback_epoch:
                    break

                # Skip PR merges as they only yield empty commits for us.
                # TODO: Find a better way to check this. Diffs don't work.
                if commit.committer.name == "GitHub" or not len(commit.parents):
                    continue

                for i in commit.diff(commit.parents[0]):
                    # If any file in this commit is a valid candidate we keep the whole commit so we check them all
                    if not self.is_valid_candidate_path(i.a_path, repo_name):
                        continue

                    if digest not in self.candidate_commits:
                        self.candidate_commits[digest] = {
                            "branched_commit": None, "original_commit": None
                        }

                    self.candidate_commits[digest][commit_key] = commit

                    # This chunk is just to make sure we don't re-apply the same commits repeatedly. It looks for
                    # a token in the branched repo commits like this ">>3908512a0bd425de80222f3c4f64b65f7af4e7d7<<"
                    # which indicates a previous graft of that commit hash from the original repo.
                    #
                    # When we get to the original repo, if we run into those commits we can apply the actual commit
                    # hash in the branched repo to self.candidate_commits, which will exclude those commits from
                    # being cloned later.
                    if repo_name == "branched":
                        previous_graft_artifact = re.search(SHA_MESSAGE_TOKEN, commit.message)

                        if previous_graft_artifact:
                            found_grafted_commits[previous_graft_artifact.groups()[1]] = commit.hexsha
                    elif self.candidate_commits[digest]["branched_commit"] is None:
                        if commit.hexsha in found_grafted_commits:
                            self.candidate_commits[digest]["branched_commit"] = found_grafted_commits[commit.hexsha]

                    # Once we've stored off the commit, no need to check more files
                    break

    def try_map_path(self, source):
        """
        Find a path in the branched repo that best matches a given one in the original repo
        :param source: A path in the original repo
        :return: The best-matching path in the branched repo, or None if no mapping is found
        """
        match = None
        match_path = ""

        for tracked_src_path in self.tracked:
            if source.startswith(tracked_src_path):
                # Assumption here is that a longer path that matches is more specific, and so potentially better
                if not match or len(tracked_src_path) > len(match):
                    # Strip out the original base path and any lingering separators
                    extra = os.path.normpath(source.replace(tracked_src_path, "", 1).lstrip("/\\"))

                    # Make a new base path for the new repo, including any subdirs
                    check_path = os.path.join(self.branched_repo_path, self.tracked[tracked_src_path], extra)

                    if os.path.exists(check_path):
                        match = tracked_src_path
                        match_path = check_path
                    else:
                        self.vprint("{} does not exist in branched repo, skipping.".format(check_path))
            else:
                self.vprint("{} not in tracked path {}".format(source, tracked_src_path))

        return match_path if match else None

    def clone_commits(self):
        """
        Copy the files of interest from a commit in the original repo to the branched repo and commit them there.
        """
        # Create a new branch in the branched repo
        if not self.dry_run:
            new_branch_name = "graft-%s" % datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
            self.output_branch = self.branched_repo.create_head(new_branch_name)
            self.output_branch.checkout()

        # Move over every applicable commit from the old repo to the new branch, moving in chronological order
        # Extra list() call here is to work around reversed() not working with OrderedDicts in Python 3.4
        # https://bugs.python.org/issue19505
        for commit_dict in reversed(list(self.candidate_commits.values())):
            original_commit = commit_dict["original_commit"]

            # We only care about commits to the original branch that hadn't made it to the new branch
            if commit_dict["original_commit"] is None or commit_dict["branched_commit"] is not None:
                continue

            # Files to be included in this commit
            found_files = []

            # Loop through all of the files in the diff, see if any files are relevant and update them
            for i in original_commit.diff(original_commit.parents[0]):
                if not self.is_valid_candidate_path(i.a_path, "original"):
                    continue

                self.vprint("-----------------------------------------------")
                self.vprint("\nFiles in commit {}".format(original_commit.hexsha))
                self.vprint(original_commit.stats.files, pretty=True)

                dest_path = self.try_map_path(i.a_path)

                if dest_path:
                    self.vprint("Copying {} to {}".format(i.a_path, dest_path))

                    if not self.dry_run:
                        with io.open(dest_path, 'wb') as f:
                            i.a_blob.stream_data(f)

                    found_files.append(dest_path)
                else:
                    self.unmatched_original_files.add(i.a_path)

            if len(found_files):
                fdate = datetime.datetime.fromtimestamp(original_commit.committed_date).isoformat()
                subject = original_commit.message.split("\n")[0].strip()
                msg = """{subject}

Grafting commit >>{sha}<< from {original_repo_name}
Original commit by {committer_name} on {date} with this message:
----------------------------------------------------------------
{message}""".format(sha=original_commit.hexsha, committer_name=original_commit.committer.name, date=fdate,
                    message=original_commit.message, subject=subject, original_repo_name=self.original_repo_name)

                print(msg)
                print("\n".join(found_files))
                print("\n\n")

                # if not self.dry_run:
                #    self.branched_repo.index.add(found_files)
                #    self.branched_repo.index.commit(msg)
            else:
                self.vprint("\n\nNothing to commit for original repo sha {}".format(original_commit.hexsha))

        if not self.dry_run:
            print("--------------------------------------------------")
            print("Your branch has been changed to {}".format(self.output_branch))
            print("--------------------------------------------------")

    def is_valid_candidate_path(self, path, repo_name):
        """
        When checking for candidate commits, see if a particular path is tracked and not ignored
        :param path: Relative path to check
        :param repo_name: "original" or "branched", used to locate tracked and ignored paths
        :return: True if the path is valid, otherwise False
        :raises AttributeError if an unknown repo name is passed in
        """
        if repo_name == "original":
            tracked = self.tracked.keys()
            ignore = self.original_ignore
        elif repo_name == "branched":
            tracked = self.tracked.values()
            ignore = self.branched_ignore
        else:
            raise AttributeError("Invalid repo name!")

        if any(path.startswith(i) for i in ignore):
            return False
        return any(path.startswith(i) for i in tracked)

    def vprint(self, string, pretty=False):
        """
        Prints a string only if it is verbose
        :param string: String to print
        :param pretty: Use prettyprint
        """
        if self.verbose:
            if pretty:
                self.pprint.pprint(string)
            else:
                print(string)

    def report(self):
        """
        Prints a short report with some information about the current state of the graft. Only really useful after
        find_candidate_commits and clone_commits have been called.
        """
        print("\nDebug Report:\n-----------------------------------------------")
        # TODO: Various other print statements should probably be moved in here and formatted correctly
        problematic_commits = set()

        for digest in self.candidate_commits:
            commit_dict = self.candidate_commits[digest]

            if commit_dict['original_commit'] is not None and commit_dict['branched_commit'] is not None:
                print("{} - exists in both repos!".format(digest))
            elif commit_dict['branched_commit'] is not None:
                print("{} - commit {} only in branched repo".format(digest, commit_dict['branched_commit']))
            elif commit_dict['original_commit'] is not None:
                print("{} - commit {} only in original repo".format(digest, commit_dict['original_commit']))
                problematic_commits.add(commit_dict["original_commit"])

        print("\nCommits in original repo that were checked against:")
        self.pprint.pprint(problematic_commits)

        print("\n{} modified files in tracked paths that do not exist in branched".format(
            len(self.unmatched_original_files)))
        self.pprint.pprint(self.unmatched_original_files)


def process_config_str(config_str):
    """
    Takes potentially multi-line RawConfigParser-returned strings, strips them, and splits them by line.
    :param config_str: String parsed in by RawConfigParser
    :return: List of strings broken up and trimmed.
    """
    if config_str is not None:
        return [i.strip() for i in config_str.split("\n") if len(i.strip()) > 0]


def process_tracked_config_str(config_lst):
    """
    Takes a list generated by process_config_str and does transformations on it to make a dict of tracked paths
    :param config_lst: List of tracked path configuration lines such as "/foo/bar/baz > /bing/bang/bar/baz"
    :return: Dict of tracked paths in the form of {"/foo/bar/baz": "/bing/bang/bar/baz"}
    """
    tracked = {}
    for line in process_config_str(config_lst):
        try:
            source, dest = line.split(">")
            source = os.path.normpath(source.strip())
            dest = os.path.normpath(dest.strip())
            tracked[source] = dest
        except:
            raise InputException("Tracked paths configured incorrectly. Each line should look have a source path, "
                                 "relative to the original repo root, a right angle bracket (>), and a destination"
                                 "path relative to the branched repo root:"
                                 "source > dest")
    return tracked


@click.command()
@click.argument('conf', type=click.Path(exists=True))
@click.option('--dry_run', is_flag=True, help="Do a test run without creating a branch or commits")
@click.option('--verbose', is_flag=True, help="Verbose output")
def main(conf, dry_run, verbose):
    """
    Creates a "best-guess" copy of commits across two unrelated (no consistent history) github repositories
    """
    config = configparser.RawConfigParser()
    config.read(conf)

    orig_repo_name = config.get("repositories", "original_repository_name")
    orig = config.get("repositories", "original_repository")
    branched = config.get("repositories", "branched_repository")
    original_branch = process_config_str(config.get("repositories", "original_branch"))[0]
    branched_branch = process_config_str(config.get("repositories", "branched_branch"))[0]

    tracked = process_tracked_config_str(config.get("tracked_paths", "tracked"))
    original_ignored = process_config_str(config.get("tracked_paths", "original_ignored"))
    branched_ignored = process_config_str(config.get("tracked_paths", "branched_ignored"))

    grafter = Grafter(orig_repo_name,
                      orig,
                      branched,
                      tracked=tracked,
                      original_ignore=original_ignored,
                      branched_ignore=branched_ignored,
                      original_repo_head=original_branch,
                      branched_repo_head=branched_branch,
                      dry_run=dry_run,
                      verbose=verbose)
    grafter.find_candidate_commits()
    grafter.clone_commits()

    if verbose:
        grafter.report()

        if not dry_run:
            print("---------------------------------------------------")
            print("WARNING: Your active git branches may have changed!")
            print("---------------------------------------------------")

if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



import pytest
import re
import six

class OEP2(object):
    def check_does_openedx_yaml_exist(self, openedx_yaml):
        assert openedx_yaml is not None

    def check_is_owner_set_in_openedx_yaml(self, openedx_yaml):
        if openedx_yaml is None:
            pytest.xfail("No openedx.yaml file found")

        if not openedx_yaml.get('archived', False) == True:
            assert 'owner' in openedx_yaml and openedx_yaml.get('owner') != "MUST FILL IN OWNER"

    def check_is_nick_set_in_openedx_yaml(self, openedx_yaml):
        if openedx_yaml is None:
            pytest.xfail("No openedx.yaml file found")

        assert 'nick' in openedx_yaml

    def check_are_tags_set_in_openedx_yaml(self, openedx_yaml):
        if openedx_yaml is None:
            pytest.xfail("No openedx.yaml file found")

        assert 'tags' in openedx_yaml

    def check_are_oeps_properly_formatted_in_openedx_yaml(self, openedx_yaml):
        if openedx_yaml is None:
            pytest.xfail("No openedx.yaml file found")

        assert 'oeps' in openedx_yaml

        for key, value in openedx_yaml['oeps'].items():
            assert re.match(r'oep-\d+', key)
            assert isinstance(value, dict) or isinstance(value, bool)

            if isinstance(value, dict):
                if value.get('applicable', True):
                    assert 'state' in value
                    assert 'reason' in value
                    assert isinstance(value['state'], bool)
                    assert isinstance(value['reason'], six.string_types)



import sys

from typing import Optional
from bowler import Query, LN, Capture, Filename, TOKEN, SYMBOL
from fissix.pytree import Node, Leaf
from lib2to3.fixer_util import Name, KeywordArg, Dot, Comma, Newline, ArgList


def filter_print_string(node, capture, filename) -> bool:
    function_name = capture.get("function_name")
    from pprint import pprint

    pprint(node)
    pprint(capture)
    return True


def filter_has_no_on_delete(node: LN, capture: Capture, filename: Filename) -> bool:
    arguments = capture.get("function_arguments")[0].children
    for arg in arguments:
        if arg.type == SYMBOL.argument and arg.children[0].type == TOKEN.NAME:
            arg_name = arg.children[0].value

            if arg_name == "on_delete":
                return False  # this call already has an on_delete argument.
    return True


def add_on_delete_cascade(
    node: LN, capture: Capture, filename: Filename
) -> Optional[LN]:
    arguments = capture.get("function_arguments")[0]
    new_on_delete_node = KeywordArg(Name(" on_delete"), Name("models.CASCADE"))

    if isinstance(arguments, Leaf): # Node is a leaf and so we need to replace it with a list of things we want instead.
        arguments.replace([arguments.clone(),Comma(),new_on_delete_node])
    else: 
        arguments.append_child(Comma())
        arguments.append_child(new_on_delete_node)

    return node


(
    Query(sys.argv[1])
    .select_method("ForeignKey")
    .is_call()
    .filter(filter_has_no_on_delete)
    .modify(add_on_delete_cascade)
    .idiff()
),
(
    Query(sys.argv[1])
    .select_method("OneToOneField")
    .is_call()
    .filter(filter_has_no_on_delete)
    .modify(add_on_delete_cascade)
    .idiff()
)




"""Show the hooks in an organization."""

import os.path
import re

import click
from git.repo.base import Repo

from edx_repo_tools.auth import pass_github
from helpers import paginated_get

@click.command()
@click.argument('org')
@click.argument('pattern', required=False)
@pass_github
def main(hub, org, pattern=None):
    for repo in hub.organization(org).repositories():
        printed_repo = False
        url = "https://api.github.com/repos/{name}/hooks".format(name=repo.full_name)
        for r in paginated_get(url):
            if pattern:
                show_it = False
                for v in r['config'].values():
                    if re.search(pattern, v):
                        show_it = True
            else:
                show_it = True

            if show_it:
                if not printed_repo:
                    print("\n-- {} ---------------------".format(repo.full_name))
                    print("  https://github.com/{}/settings/hooks".format(repo.full_name))
                    printed_repo = True
                print("{r[name]}".format(r=r))
                for k, v in sorted(r['config'].items()):
                    print("  {k}: {v}".format(k=k, v=v))



from setuptools import setup


with open('README.rst') as readme:
    long_description = readme.read()

setup(
    name='edx-repo-tools',
    version='0.2.8',
    description="This repo contains a number of tools Open edX uses for working with GitHub repositories.",
    long_description=long_description,
    license='Apache',
    keywords='edx repo tools',
    url='https://github.com/edx/repo-tools',
    author='edX',
    author_email='oscm@edx.org',
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License'
    ],
    packages=[
        'edx_repo_tools',
    ],
    install_requires=[
        'appdirs',
        'click',
        'lazy',
        'github3.py',
        'pytest',
        'pytest-xdist',
        'pyyaml',
        'ruamel.yaml'
    ],
    entry_points={
        'console_scripts': [
            'clone_org = edx_repo_tools.dev.clone_org:main',
            'show_hooks = edx_repo_tools.dev.show_hooks:main',
            'oep2 = edx_repo_tools.oep2:_cli',
            'sync_labels = edx_repo_tools.ospr.sync_labels:sync_labels',
            'no_yaml = edx_repo_tools.ospr.no_yaml:no_yaml',
            'tag_release = edx_repo_tools.release.tag_release:main',
            'drip = edx_repo_tools.drip_survey:cli',
            'get_org_repo_urls = edx_repo_tools.dev.get_org_repo_urls:main',
            'modernize_travis = django3_codemods.config_tools.travis_modernizer:main',
            'modernize_tox = django3_codemods.config_tools.tox_modernizer:main',
            'modernize_openedx_yaml = edx_repo_tools.modernize_openedx_yaml:main',
            'modernize_github_actions = django3_codemods.config_tools.github_actions_modernizer:main',
            'add_common_constraint = edx_repo_tools.add_common_constraint:main',
        ],
    },
    package_data={
        'edx_repo_tools.oep2.report': ['oep2-report.ini'],
    }
)



"""
``oep2 report``: Check repositories for OEP-2 compliance.
"""

import logging
import pkg_resources
import sys

import click
from git.cmd import Git
import pytest

from edx_repo_tools.auth import pass_github
from .plugin import Oep2ReportPlugin

LOGGER = logging.getLogger(__name__)


@click.command(context_settings={'ignore_unknown_options': True})
@click.option(
    '--trace/--no-trace', default=False,
    help="Trace git and github interactions during reporting",
)
@click.argument(
    'pytest_args',
    nargs=-1,
    type=click.UNPROCESSED,
)
@pass_github
def cli(hub, trace, pytest_args):
    """
    Command-line interface specification for ``oep2 report``.
    """
    args = [
        '--pyargs', 'edx_repo_tools.oep2.checks',
        '-c', pkg_resources.resource_filename(__name__, 'oep2-report.ini'),
    ]

    if trace:
        args.extend(['-s', '-vvv'])
        Git.GIT_PYTHON_TRACE = True
    else:
        args.append('-q')

    args.extend(pytest_args)

    plugins = [
        Oep2ReportPlugin(hub),
    ]

    sys.exit(pytest.main(args=args, plugins=plugins))



import click
from ruamel.yaml import YAML


def dry_echo(dry, message, *args, **kwargs):
    """
    Print a command to the console (like :func:`click.echo`), but if ``dry`` is True,
    then prefix the message with a warning message stating that the action was
    skipped. All unknown args and kwargs are passed to :func:`click.echo`

    Example usage:

        dry_echo(dry, "Firing ze missiles!", fg='red')
        if not dry:
            fire_ze_missiles()

    Arguments:
        dry (bool): Whether to prefix the dry-run notification
        message: The message to print
    """
    click.echo("{dry}{message}".format(
        dry=click.style("DRY RUN - SKIPPED: ", fg='yellow', bold=True) if dry else "",
        message=click.style(message, *args, **kwargs)
    ))


def dry(f, help='Disable or enable actions taken by the script'):
    """
    A click decorator that adds a ``--dry/--doit`` flag. It is passed to the
    command as ``dry``, and defaults to True.
    """
    return click.option(
        '--dry/--doit',
        is_flag=True,
        default=True,
        help=help,
    )(f)


class YamlLoader:
    def __init__(self, file_path):
        self.file_path = file_path
        self.yml_instance = YAML()
        self.yml_instance.indent(mapping=2, sequence=4, offset=2)
        self._load_file()

    def _load_file(self):
        with open(self.file_path, 'r') as file_stream:
            self.elements = self.yml_instance.load(file_stream)

    def update_yml_file(self):
        with open(self.file_path, 'w') as file_stream:
            self.yml_instance.dump(self.elements, file_stream)



"""
Top-level definition of the ``oep2`` commandline tool.
"""

import click

from . import explode_repos_yaml
from .report.cli import cli as report_cli


def _cli():
    cli(auto_envvar_prefix="OEP2")


@click.group()
def cli():
    """
    Tools for implementing and enforcing OEP-2.
    """
    pass

cli.add_command(explode_repos_yaml.implode)
cli.add_command(report_cli, 'report')



"""
``oep2 implode``: aggregate many openedx.yaml files.
"""

import click
from github3.exceptions import NotFoundError
import logging
import yaml

from edx_repo_tools.auth import pass_github
from edx_repo_tools.data import iter_openedx_yaml


logging.basicConfig()
LOGGER = logging.getLogger(__name__)


@click.command()
@pass_github
@click.option('--org', multiple=True, default=['edx', 'edx-ops', 'edx-solutions',])
@click.option(
    '--branch',
    multiple=True,
    default=None,
    help="The branch(es) to examine for openedx.yaml files. If more than one, "
         "the first found will be used."
)
def implode(hub, org, branch):
    """
    Implode all openedx.yaml files, and print the results as formatted output.
    """
    data = {
        repo.full_name: openedx_yaml
        for repo, openedx_yaml
        in iter_openedx_yaml(hub, org, branch)
    }
    click.echo(yaml.safe_dump(data, encoding=None, indent=4))
